from pytest import mark
from pytest_django.asserts import assertHTMLEqual

from django_mistune.templatetags.mistune import markdown

ALL_STYLES = ("style", ("headers", "classes", "links"))


@mark.parametrize(*ALL_STYLES)
def test_empty_string(style):
    assertHTMLEqual(markdown("", style), "")


def test_header_level():
    assertHTMLEqual(markdown("# Test", "headers"), "<h3>Test</h3>")


def test_header_level_nested():
    assertHTMLEqual(
        markdown("- Hello\n\n - World\n\n - ## Heading", "headers"),
        """<ul>
            <li><p>Hello</p></li>
            <li><p>World</p></li>
            <li><h4>Heading</h4></li>
        </ul>""",
    )


def test_classes():
    assertHTMLEqual(
        markdown("Hello *World*", "classes"),
        '<p class="a b">Hello <em class="c">World</em></p>',
    )


def test_classes_existing_attrs():
    assertHTMLEqual(
        markdown("This ![image](test.jpg)", "classes"),
        '<p class="a b">This <img src="test.jpg" alt="image" class="d"></p>',
    )


def test_links():
    target = '<p><a target="_blank" href="example.com">Link</a></p>'
    assertHTMLEqual(markdown("[Link](example.com)", "links"), target)
    assertHTMLEqual(markdown("[Link][1]\n\n[1]: example.com", "links"), target)



from django.template import Context, Template
from pytest import fixture
from pytest_django.asserts import assertHTMLEqual


def assert_renders(template, result, **kwargs):
    """Small utility function to render Django templates quickly."""
    assertHTMLEqual(template.render(Context(kwargs)), result)


@fixture
def default_template():
    return Template(
        r"""{% load mistune %}
        {{ text|markdown }}"""
    )


@fixture
def noescape_template():
    return Template(
        r"""{% load mistune %}
        {% autoescape off %}
        {{ text|markdown }}
        {% endautoescape %}"""
    )


def test_simple_render(default_template):
    assert_renders(
        default_template,
        "<p><strong>Hello</strong> <em>World</em></p>",
        text="**Hello** *World*",
    )


def test_escape(default_template):
    assert_renders(default_template, "<p>&lt;br&gt;</p>", text="<br>")


def test_noescape(noescape_template):
    assert_renders(noescape_template, "<br>", text="<br>")



from pytest import raises
from pytest_django.asserts import assertHTMLEqual

from django_mistune.conf import settings
from django_mistune.templatetags.mistune import markdown
from django_mistune.utils import get_style


def test_read_settings():
    assert "noescape" in settings.MISTUNE_STYLES
    assert "default" in settings.MISTUNE_STYLES


def test_get_style_unknown():
    with raises(ValueError):
        get_style("unknown")


def test_get_style():
    assert not get_style("noescape")["escape"]


def test_simple_gen():
    assertHTMLEqual(
        markdown("**Hello** *World*"),
        "<p><strong>Hello</strong> <em>World</em></p>",
    )


def test_html_escape():
    escaped = "<p>&lt;br&gt;</p>"
    assertHTMLEqual(markdown("<br>"), escaped)
    assertHTMLEqual(markdown("<br>", "noescape"), escaped)



from types import SimpleNamespace

from django.conf import settings as dj_settings

settings = SimpleNamespace(
    # A style is basically a dict containing parameters to pass to `mistune`
    MISTUNE_STYLES={
        **{
            "default": {
                # The `escape` parameter is useless as it will be overriden
                # according to Django `autoescape`
                "hard_wrap": False,
                "renderer": "html",
                "plugins": [],
            },
        },
        **getattr(dj_settings, "MISTUNE_STYLES", {}),
    }
)



from django_mistune.plugins import AddClasses, HeaderLevels, TargetBlankLinks

USE_TZ = True  # Avoid deprecation warning

INSTALLED_APPS = [
    "django_mistune",
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
    },
]

MISTUNE_STYLES = {
    "noescape": {"escape": False},
    "headers": {"plugins": [HeaderLevels(3)]},
    "classes": {
        "plugins": [AddClasses({"p": ("a", "b"), "em": "c", "img": "d"})]
    },
    "links": {"plugins": [TargetBlankLinks()]},
}



import mistune
from django import template
from django.template.defaultfilters import stringfilter
from django.utils.safestring import mark_safe

from django_mistune.utils import get_style

register = template.Library()


@register.filter(needs_autoescape=True)
@stringfilter
def markdown(text, style_name="default", autoescape=True):
    """Converts `text` to HTML. If a `style_name` is given, it will be loaded
    from the settings.

    The `escape` parameter of the style will always be ignored as it will be
    overriden by the `autoescape` behavior of Django."""
    style = get_style(style_name)
    markdown = mistune.create_markdown(**{**style, "escape": autoescape})

    return mark_safe(markdown(text))



from django_mistune.conf import settings


def get_style(style_name):
    style = settings.MISTUNE_STYLES.get(style_name)
    if style is None:
        raise ValueError(
            f"Unknown style {style_name}. "
            f"Avalaible styles: {', '.join(settings.MISTUNE_STYLES)}"
        )
    return style



from defusedxml import ElementTree
from defusedxml.ElementTree import ParseError


class HeaderLevels:
    def __init__(self, top=3):
        """Demotes all the headers of the document so the first level becomes
        the `top` parameter.

        This changes both the AST and the rendering of the document.
        """
        self.top = top
        assert top > 0, "Top level must be >= 1"

    def replace_level(self, token):
        """Replace the level of a heading token."""
        level = token["params"][0]
        token["params"] = (level + self.top - 1,)

    def find_headings(self, md, tokens, state):
        """Walk the token tree and search for heading tokens."""
        for tok in tokens:
            if tok["type"] == "heading":
                self.replace_level(tok)
            if "children" in tok.keys():
                self.find_headings(md, tok["children"], state)
        return tokens

    def __call__(self, md):
        md.before_render_hooks.append(self.find_headings)


class AddClasses:
    def __init__(self, tags=None):
        """Add HTML classes to all given HTML tags.

        For example, `tags={'p': ('a', 'b'), 'em': 'a'}` will add the `a` class
        to `p` and `em` tags of the document, and the `b` class to `p` tags.

        This only alters the rendering of the document, not its AST.

        If the input HTML cannot be parsed, it will be returned as-is.
        """
        self.tags = tags or {}

    def add_classes(self, md, result, state):
        """Find bare HTML tags and add classes to them."""
        try:
            root = ElementTree.fromstring(result)
        except ParseError:
            return result

        for tag, classes in self.tags.items():
            if not isinstance(classes, tuple):
                classes = (classes,)

            self._process_tag(root, tag, classes)

        return ElementTree.tostring(root, encoding="unicode")

    def _process_tag(self, root, tag, new_classes):
        """Add `new_classes` to all `tag` elements in `root`."""
        for el in root.iter(tag):
            old_classes = el.get("class", "").split(" ")
            classes = set((*old_classes, *new_classes))

            el.set("class", " ".join(classes))

    def __call__(self, md):
        if md.renderer.NAME == "html":
            md.after_render_hooks.append(self.add_classes)


class TargetBlankLinks:
    """Add `target="_blank"` to all document links.

    This only alters the rendering of the document, not its AST.

    If the input HTML cannot be parsed, it will be returned as-is.
    """

    def add_targets(self, md, result, state):
        """Find bare `<a>` tags and add targets to them."""
        try:
            root = ElementTree.fromstring(result)
        except ParseError:
            return result

        for el in root.iter("a"):
            el.set("target", "_blank")

        return ElementTree.tostring(root, encoding="unicode")

    def __call__(self, md):
        if md.renderer.NAME == "html":
            md.after_render_hooks.append(self.add_targets)
